package jp.hishidama.afwhs
import jp.hishidama.afwhs.prove.Node
import scala.collection.mutable.LinkedHashMap
import jp.hishidama.afwhs.gen.DmdlGen
import jp.hishidama.afwhs.gen.OperatorGen
import jp.hishidama.afwhs.gen.FlowGen
import jp.hishidama.afwhs.gen.ImporterGen
import jp.hishidama.afwhs.gen.ExporterGen

class Generator(val workspace: String, val project: String, val basePackage: String, val baseName: String) extends AnyRef with Util {
  var operatorPackage = basePackage + ".operator"
  var operatorClass = baseName + "Operator"
  def operatorFactoryPackage = basePackage + ".operator"
  def operatorFactoryClass = operatorClass + "Factory"
  var flowPackage = basePackage + ".jobflow"
  var flowClass = baseName + "Job"
  var gatewayPackage = flowPackage + ".gateway"

  def newSource[A <: Model: ClassManifest](): Source[A] = {
    val s = Source[A]()
    s.gen = this
    ins :+= s
    s
  }
  def out(r: Source[_]) {
    r.gen = this
    outs :+= r
  }
  def out(rs: Product) {
    rs.productIterator.foreach { r =>
      out(r.asInstanceOf[Source[_]])
    }
  }

  var ins: Seq[Source[_]] = Seq.empty
  var outs: Seq[Source[_]] = Seq.empty
  var nodes: Seq[Node] = Seq.empty

  def addNode(node: Node) = {
    nodes :+= node
    node.ss.foreach { s => s.gen = this }
    node.rs.foreach { r => r.gen = this }
    this
  }

  override def toString(): String = {
    "Generator{\nin =" + ins + "\nout=" + outs + "\n" + nodes.mkString("\n") + "\n}"
  }

  import java.io._

  def generateDmdl(fileName: String) {
    generateDmdl(new File(new File(workspace, project + "/src/main/dmdl"), fileName))
  }
  def generateDmdl(f: File) {
    println("generateDmdl:\t" + f)
    val writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(f), "UTF-8"))
    try {
      dmdlMap.values.foreach { _.generate(writer) }
    } finally {
      writer.close()
    }
  }

  def generateOperator() {
    val gen = new OperatorGen(this)

    val f = new File(new File(workspace, project + "/src/main/java/" + operatorPackage.replaceAll("\\.", "/")), operatorClass + ".java")
    println("generateOperator:\t" + f)
    val writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(f), "UTF-8"))
    try {
      gen.generate(writer)
    } finally {
      writer.close()
    }
  }

  def generateImporter() {
    val gen = new ImporterGen(this)
    gen.packageName = basePackage + ".jobflow.gateway"
    ins.foreach { s =>
      val f = new File(new File(workspace, project + "/src/main/java/" + gen.packageName.replaceAll("\\.", "/")), s.modelClassName + "Importer.java")
      if (!f.exists()) {
        f.getParentFile().mkdirs()
        println("generateImporter:\t" + f)
        val writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(f), "UTF-8"))
        try {
          gen.generate(writer, s)
        } finally {
          writer.close()
        }
      }
    }
  }

  def generateExporter() {
    val gen = new ExporterGen(this)
    gen.packageName = basePackage + ".jobflow.gateway"
    outs.foreach { s =>
      val f = new File(new File(workspace, project + "/src/main/java/" + gen.packageName.replaceAll("\\.", "/")), s.modelClassName + "Exporter.java")
      if (!f.exists()) {
        f.getParentFile().mkdirs()
        println("generateExporter:\t" + f)
        val writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(f), "UTF-8"))
        try {
          gen.generate(writer, s)
        } finally {
          writer.close()
        }
      }
    }
  }

  def generateFlow() {
    val gen = new FlowGen(this)
    val f = new File(new File(workspace, project + "/src/main/java/" + flowPackage.replaceAll("\\.", "/")), flowClass + ".java")
    if (!f.exists() || true) {
      println("generateFlow:\t" + f)
      val writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(f), "UTF-8"))
      try {
        gen.generate(writer)
      } finally {
        writer.close()
      }
    }
  }

  lazy val dmdlMap = {
    val map = LinkedHashMap.empty[String, DmdlGen]
    nodes.foreach { n =>
      n.ss.foreach { s =>
        val name = createInstance(s.c0).modelName
        map.getOrElseUpdate(name, new DmdlGen(name, s))
      }
      n.rs.foreach { s =>
        val name = createInstance(s.c0).modelName
        map.getOrElseUpdate(name, new DmdlGen(name, s)).addNode(n)
      }
    }
    map
  }
}

private[afwhs] trait WithGenerator {
  var gen: Generator = _
}
